
# require 'rails'
require 'ext_logger'
# require 'ext_batch/common'
require 'sys/proctable' if !defined? ProcTable

class ExtBatch
  include Sys
  include ExtLogger::Common
  # include Common

  LABEL_OK    = 'ok'
  LABEL_ERROR = 'error'
  LABEL_WARN  = 'warning'

  SEND_MAIL_NO   = 'no'
  SEND_MAIL_YES  = 'yes'

  PID_FILE_PATH = ExtLogger.rails_root + '/tmp/batch_pids/'

  def self.monitor_pid_file(batch=nil)
    logger = batch if batch.is_a?(ExtBatch)

    path = self.get_pid_file_path
    if File.directory?(path)
      Dir.foreach(path) do |file|
        file_name = File.basename(file)
        if file_name.include?('.pid')
          pid = nil

          # Get batch pid file
          File.open(File.join(path, file_name), 'r') do |f|
            while line = f.gets
              pid = line.to_i
            end
          end

          # Delete batch pid file
          if pid && !ProcTable.ps(pid: pid)
            logger.info "Delete file: #{file_name}, pid: #{pid}"
            File.delete(file_name)
          end
        end
      end
    end
  end

  # Get file name for batch process pid
  def self.get_pid_file_name
    return @pid_file[:name] || ''
  end

  # Set file path for batch process pid
  def self.set_pid_file_path(str)
    @@pid_file_path = str if str.is_a?(String) && !is_blank?(str)
  end

  # Get file path for batch process pid
  def self.get_pid_file_path
    return @@pid_file_path || PID_FILE_PATH
  end

  # Set file name for batch process pid
  def self.set_pid_file_name(str)
    @pid_file[:name] = str if str.is_a?(String) && !is_blank?(str)
  end

  # Set file name for batch process pid
  def set_pid_flag(flag=true)
    flag = true if !flag.is_a?(TrueClass) && !flag.is_a?(FalseClass)
    @pid_flag = flag
  end

  # Initialize, next is parameters:
  # 1rst: task object or batch name(string) (Default parameter)
  # 2end: option (Influencing log size, file name format, etc.)
  #   :log            -- Inheriting ExtLogger classes, hash class, value exp:
  #     {
  #       :age            -- Split file with N days
  #       :size           -- Split file with file size
  #       :split_format   -- Split file format. exp: '%Y....'
  #       :log_max_num    -- Number of log files saved
  #       :debug          -- Is debug
  #     }
  def initialize(*args)
    func_name = "[#{get_class_name}.#{__method__.to_s}]"
    output_debug "#{func_name} args.size: #{args.size}"

    # Instance variables
    @task_name      = ''
    @ext_logger     = nil
    @batch_log_name = ''
    @batch_message  ||= ''
    @start_time     ||= Time.now
    @last_time      ||= @start_time

    @pid_flag = true
    @@pid_file_path ||= PID_FILE_PATH # Prevent duplication after file sharing by multiple servers
    @pid_file = {
        name: '',
        cmd: ''
    }

    init(args[0]) if args.size == 1

    if args.size == 2
      opt = args[1].clone
      if opt.is_a?(Hash)
        init(args[0], opt)
      else
        init(args[0])
      end
    end
  end

  # Get the message
  def get_message
    return @batch_message
  end

  # Set the message
  def set_message(str=nil)
    @batch_message = str if !is_blank?(str)
  end

  # Record start time
  def start_time
    return @start_time
  end

  # Real-time display of usage time
  def show_use_time
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    now_time = Time.now

    output_debug "#{func_name} @start_time:#{@start_time}, @last_time:#{@last_time}, now_time:#{now_time}"

    use_time = format_sub_time(@last_time, now_time)
    message = "Use time: #{use_time}#{get_total_time(@start_time, now_time)}"
    @ext_logger.info message

    @last_time = now_time
  end

  # batch end
  def end(opt={})
    func_name = "[#{get_instance_name}.end]"
    message = "#{@task_name} END."
    # display usage time
    # opt[:show_use_time] = true if opt[:show_use_time] != false
    message += get_total_time(@start_time) #if opt[:show_use_time]

    delete_pid_file if @pid_flag
    @ext_logger.info message
    @ext_logger.record_memory('end') if defined?(@ext_logger.record_memory)
  end

  # Inheriting ExtLogger instance method: info
  def info(*args)
    @ext_logger.info(*args)
  end

  # Inheriting ExtLogger instance method: info
  def debug(*args)
    @ext_logger.debug(*args)
  end

  # Inheriting ExtLogger instance method: info
  def error(*args)
    @ext_logger.error(*args)
  end

  # Inheriting ExtLogger instance method: info
  def warn(*args)
    @ext_logger.warn(*args)
  end

  # Inheriting ExtLogger instance method: info
  def fatal(*args)
    @ext_logger.fatal(*args)
  end

  # Inheriting ExtLogger instance method: info
  def exception(e, opt={})
    opt = {} if !opt.is_a?(Hash)
    opt[:send_mail] ||= SEND_MAIL_NO
    opt[:status]    = LABEL_ERROR
    content = []
    content << "#{@log_name} #{opt[:status]} (#{Time.now})"
    content << analysis_exception(e)
    content << opt[:content] if !is_blank?(opt[:content])

    mail_error({content: content.join("\n")}) if opt[:send_mail] == SEND_MAIL_YES
    error(content)
  end

  # batch: send ok mail
  def mail_ok(opt={})
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    opt = {} if !opt.is_a?(Hash)
    opt[:content] ||= ''
    opt[:status] = LABEL_OK
    return mail_tips(@log_name, opt)
  end

  # batch: send error mail
  def mail_error(opt={})
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    opt = {} if !opt.is_a?(Hash)
    opt[:content] ||= ''
    opt[:status] = LABEL_ERROR
    return mail_tips(@log_name, opt)
  end

  # batch: send warning mail
  def mail_warn(opt={})
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    opt = {} if !opt.is_a?(Hash)
    opt[:content] ||= ''
    opt[:status] = LABEL_WARN
    return mail_tips(@log_name, opt)
  end

  # batch: send exception mail
  def mail_exception(e, opt={})
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    opt[:status] = LABEL_ERROR
    content = []
    content << "#{@log_name} #{opt[:status]} (#{Time.now})"
    content << analysis_exception(e)
    content << opt[:content] if !is_blank?(opt[:content])
    opt[:content] = content.join("\n")

    return mail_tips(@log_name, opt)
  end

  private
  # Private function for initialize
  # batch initialize
  def init(task, opt={})
    require 'rake'
    func_name = "[#{get_class_name}:private #{__method__.to_s}]"
    set_debug(true) if opt[:debug]

    output_debug "#{func_name} task_name: #{task}, opt: #{opt}"

    # Get log name & task name
    if task.is_a?(Rake::Task)
      @log_name = task.name

    elsif task.is_a?(String)
      @log_name = task

    else
      @log_name = "ext_batch_log_#{Time.now.to_i}"
    end
    @task_name      = @log_name.clone
    @batch_message  = @log_name.clone

    # Format log path & name
    @log_name = @log_name.gsub(':', '/')
    while @log_name.include?('//')
      @log_name.gsub!('//', '/')
    end
    if @log_name[0] == '/'
      @log_name = @log_name[1, @log_name.size - 1]
    end

    # Full log file name
    log_path = ExtLogger.rails_root + '/log/' + @log_name + '.log'
    mkdir_more(log_path)
    output_debug "#{func_name} Current log path: #{log_path}"

    opt[:log] ||= {}
    opt[:log][:debug] = get_debug
    @ext_logger = ExtLogger.new(log_path, log: opt[:log])
    @ext_logger.record_memory()
    @batch_message = ''

    if task.nil?
      @ext_logger.error "#{func_name} Object Error! [task] = nil"
      return

    else
      if task.class != String
        if is_blank?(task.name)
          @ext_logger.error "#{func_name} Object Error! [task] = #{task.to_json}"
          return
        end
      end
    end

    # # Additional ENV parameters to message
    # if !is_blank?(opt[:env]) && opt[:env].class == Array
    #   args = []
    #   opt[:env].each do |item|
    #     args << "#{item}=#{ENV[item]}" if !is_blank?(ENV[item])
    #   end
    #   @batch_message += "[#{args.join(',')}]" if args.size > 0
    # end

    # Additional content to message
    if opt[:message].is_a?(String) && !is_blank?(opt[:message])
      @batch_message += opt[:message]
    end

    set_pid_flag(opt[:pid_flag]) if !is_blank?(opt[:pid_flag])
    if @pid_flag
      @pid_file = {
          name: @task_name.gsub(':', '#').gsub(' ', '_') + '.pid',
          cmd:  @task_name
      }
      create_pid_file
    end

    message = "(#{ExtLogger.rails_env}) #{@task_name} START"
    @ext_logger.info message
  end

  # batch: send tips mail
  def mail_tips(task_name, opt={})
    func_name = "[#{get_class_name}:private #{__method__.to_s}]"
    opt[:from] ||= ''
    opt[:to] ||= ''
    opt[:content] ||= ''
    _status = opt[:status] || LABEL_OK

    result = []
    result << "System time: #{Time.now}, #{get_total_time(@start_time)}"
    result << ''
    result << "BATCH: #{task_name}"
    result << "STATUS: #{_status}"
    result << ''
    result << "NOTE: #{opt[:content]}" if !is_blank?(opt[:content])

    message_body = result.join("\n")
    # info message_body

    mail_option = {
        to:     opt[:to],
        title:  "BATCH #{_status}: #{task_name} (#{Time.now})",
        body:   message_body.gsub("\n", "<br />"),
        status: _status
    }

    message =   "### Next is detail of mail ###\n"
    message +=  "[TITLE]: #{mail_option[:title]}\n"
    message +=  "[BODY]:\n #{message_body}\n"

    case _status
      when LABEL_ERROR
        @ext_logger.error message
      when LABEL_OK
        @ext_logger.info message
      when LABEL_WARN
        @ext_logger.warn message
      else
        @ext_logger.info message
    end

    if defined? ActionMailer
      mail_flag = true
      if is_blank?(opt[:to]) && is_blank?(opt[:content])
        @ext_logger.error "#{func_name} Mail delivery failed, parameters error. opt: #{opt}"
        mail_flag = false
      end

      if mail_flag
        # mail_option = {
        #     from: opt[:from],
        #     to: opt[:to],
        #     content_type: 'text/html',
        #     subject: _option[:title],
        #     body: _option[:body]
        # }
        # # logger.info "set_mail", mail_hash
        # mail_option = mail_option.merge(:template_path => opt[:template_path]) if !is_blank?(opt[:template_path])
        # mail_option = mail_option.merge(:template_name => opt[:template_name]) if !is_blank?(opt[:template_name])
        # mail(mail_option)

        begin
          ApplicationMailer.send_batch_mail(mail_option[:to], mail_option[:title], mail_option[:body]).deliver
          @ext_logger.info "Batch mail send OK."
        rescue Exception => e
          @ext_logger.exception e
        end
      end
    else
      @ext_logger.warn "#{func_name} Mail delivery failed, not found ActionMailer class."
    end
  end

  # Create batch process pid file
  def create_pid_file
    func_name = "[#{get_class_name}:private #{__method__.to_s}]"
    if !is_blank?(@pid_file[:name]) && !is_blank?(@@pid_file_path)
      file_path = File.join(@@pid_file_path, @pid_file[:name])
      @ext_logger.info "#{func_name} pid_file_path: #{file_path}"

      pid = nil
      a = ProcTable.ps
      a.each_with_index do |p, idx|
        if p.cmdline.include?(@pid_file[:cmd])
          @ext_logger.info "pid:#{p.pid}, cmd:#{p.comm}, cmdline: #{p.cmdline}"
          pid = p.pid
          break
        end
        # puts "pid:#{p.pid}, cmd:#{p.comm}, cmdline: #{p.cmdline}" if p.cmdline.include?(@task_name)
        # puts "p.class: #{p.class}"
        # puts p.to_s
        # break if idx >= 10
        # ...
      end

      if pid
        @ext_logger.info "#{func_name} #{@pid_file[:name]}.pid: #{pid}"
        mkdir_more(file_path)
        File.delete(file_path) if File.exist?(file_path)

        File.open(file_path, 'w') do |f|
          f.write pid
          f.close
        end

      else
        @ext_logger.warn("#{func_name} Pid file generation failed, process[#{@pid_file[:name]}] not found.")
      end
    end
  end

  # Delete batch process pid file
  def delete_pid_file
    func_name = "[#{get_class_name}:private #{__method__.to_s}]"
    if !is_blank?(@pid_file[:name]) && !is_blank?(@@pid_file_path)
      file_path = File.join(@@pid_file_path, @pid_file[:name].gsub(' ', '_'))

      File.delete(file_path) if File.exist?(file_path)
    end
  end

  def get_total_time(s_time, e_time=Time.now)
    func_name = "[#{get_instance_name}.#{__method__.to_s}]"
    total_use_time = format_sub_time(s_time, e_time)
    return "[total use time: #{total_use_time}]"
  end

end
